@@ -25,8 +25,6 @@ Atom _xklAtoms[TOTAL_ATOMS];
 
 Window _xklRootWindow;
 
-Bool _xklAllowSecondaryGroupOnce;
-
 int _xklDefaultGroup;
 
 Bool _xklSkipOneRestore;
@@ -37,6 +35,10 @@ int _xklDebugLevel = 0;
 
 Window _xklPrevAppWindow;
 
+int _xklListenerType = 0;
+
+XklVTable *xklVTable = NULL;
+
 XklConfigCallback _xklConfigCallback = NULL;
 void *_xklConfigCallbackData;
 
@@ -52,6 +54,7 @@ static Bool groupPerApp = True;
 
 static Bool handleIndicators = False;
 
+
 void XklSetIndicatorsHandling( Bool whetherHandle )
 {
   handleIndicators = whetherHandle;
@@ -77,10 +80,53 @@ Bool XklIsGroupPerApp( void )
   return groupPerApp;
 }
 
+static void _XklSetSwitchToSecondaryGroup( Bool val )
+{
+  CARD32 propval = (CARD32)val;
+  XChangeProperty( _xklDpy, _xklRootWindow, _xklAtoms[XKLAVIER_ALLOW_SECONDARY],
+                   XA_INTEGER, 32, PropModeReplace,
+                   (unsigned char*)&propval, 1 );
+  XSync( _xklDpy, False );
+}
+
 void XklAllowOneSwitchToSecondaryGroup( void )
 {
   XklDebug( 150, "Setting allowOneSwitchToSecondaryGroup flag\n" );
-  _xklAllowSecondaryGroupOnce = True;
+  _XklSetSwitchToSecondaryGroup( True );
+}
+
+Bool _XklIsOneSwitchToSecondaryGroupAllowed( void )
+{
+  Bool rv = False;
+  unsigned char *propval = NULL;
+  Atom actualType;
+  int actualFormat;
+  unsigned long bytesRemaining;
+  unsigned long actualItems;
+  int result;
+
+  result = XGetWindowProperty( _xklDpy, _xklRootWindow, 
+                               _xklAtoms[XKLAVIER_ALLOW_SECONDARY], 0L, 1L,
+                               False, XA_INTEGER, &actualType, &actualFormat,
+                               &actualItems, &bytesRemaining,
+                               &propval );
+
+  if( Success == result )
+  {
+    if( actualFormat == 32 && actualItems == 1 )
+    {
+      rv = *(Bool*)propval;
+    }
+    XFree( propval );
+  }
+
+  return rv;
+}
+
+void _XklOneSwitchToSecondaryGroupPerformed( void )
+{
+  XklDebug( 150, "Resetting allowOneSwitchToSecondaryGroup flag\n" );
+  _XklSetSwitchToSecondaryGroup( False );
 }
 
 void XklSetDefaultGroup( int group )
@@ -129,8 +175,15 @@ void XklSetLogAppender( XklLogAppender fun )
   logAppender = fun;
 }
 
-int XklStartListen(  )
+int XklStartListen( int what )
 {
+  _xklListenerType = what;
+  
+  if( !( xklVTable->features & XKLF_REQUIRES_MANUAL_LAYOUT_MANAGEMENT ) &&
+      ( what & XKLL_MANAGE_LAYOUTS ) )
+    XklDebug( 0, "The backend does not require manual layout management - "
+                 "but it is provided by the application" );
+  
   XklResumeListen(  );
   _XklLoadWindowTree(  );
   XFlush( _xklDpy );
@@ -143,6 +196,74 @@ int XklStopListen(  )
   return 0;
 }
 
+int XklInit( Display * a_dpy )
+{
+  int scr;
+  char *sdl;
+  int rv;
+
+  sdl = getenv( "XKL_DEBUG" );
+  if( sdl != NULL )
+  {
+    XklSetDebugLevel( atoi( sdl ) );
+  }
+
+  if( !a_dpy )
+  {
+    XklDebug( 10, "XklInit : display is NULL ?\n");
+    return -1;
+  }
+
+  _xklDefaultErrHandler =
+    XSetErrorHandler( ( XErrorHandler ) _XklErrHandler );
+
+  _xklDpy = a_dpy;
+  scr = DefaultScreen( _xklDpy );
+  _xklRootWindow = RootWindow( _xklDpy, scr );
+
+  _xklSkipOneRestore = False;
+  _xklDefaultGroup = -1;
+  _xklSecondaryGroupsMask = 0L;
+  _xklPrevAppWindow = 0;
+
+  _xklAtoms[WM_NAME] = XInternAtom( _xklDpy, "WM_NAME", False );
+  _xklAtoms[WM_STATE] = XInternAtom( _xklDpy, "WM_STATE", False );
+  _xklAtoms[XKLAVIER_STATE] = XInternAtom( _xklDpy, "XKLAVIER_STATE", False );
+  _xklAtoms[XKLAVIER_TRANSPARENT] =
+    XInternAtom( _xklDpy, "XKLAVIER_TRANSPARENT", False );
+  _xklAtoms[XKLAVIER_ALLOW_SECONDARY] =
+    XInternAtom( _xklDpy, "XKLAVIER_ALLOW_SECONDARY", False );
+
+  _XklOneSwitchToSecondaryGroupPerformed();
+
+  rv = -1;
+  XklDebug( 150, "Trying all backends:\n" );
+#ifdef ENABLE_XKB_SUPPORT
+  XklDebug( 150, "Trying XKB backend\n" );
+  rv = _XklXkbInit();
+#endif
+#ifdef ENABLE_XMM_SUPPORT
+  if( rv != 0 ) 
+  {
+    XklDebug( 150, "Trying XMM backend\n" );
+    rv = _XklXmmInit();
+  }
+#endif
+  if( rv == 0 )
+  {
+    XklDebug( 150, "Actual backend: %s\n",
+              XklGetBackendName() );
+  }
+  else
+  {
+    XklDebug( 0, "All backends failed, last result: %d\n", rv );
+    _xklDpy = NULL;
+  }
+
+  return ( rv == 0 ) ?
+    ( _XklLoadAllInfo() ? 0 : _xklLastErrorCode ) : -1;
+}
+
 int XklTerm(  )
 {
   XSetErrorHandler( ( XErrorHandler ) _xklDefaultErrHandler );
@@ -156,28 +277,28 @@ int XklTerm(  )
   return 0;
 }
 
-Bool XklGrabKey( int key, unsigned modifiers )
+Bool XklGrabKey( int keycode, unsigned modifiers )
 {
-  int keyCode;
   Bool retCode;
   char *keyName;
 
-  keyCode = XKeysymToKeycode( _xklDpy, key );
-  keyName = XKeysymToString( key );
+  if( _xklDebugLevel >= 100 )
+  {
+    keyName = XKeysymToString( XKeycodeToKeysym( _xklDpy, keycode, 0 ) );
+    XklDebug( 100, "Listen to the key %d/(%s)/%d\n", 
+                   keycode, keyName, modifiers );
+  }
 
-  XklDebug( 100, "listen to the key %X(%d/%s)/%d\n", key, keyCode, keyName,
-            modifiers );
-
-  if( ( KeyCode ) NULL == keyCode )
+  if( ( KeyCode ) NULL == keycode )
     return False;
 
   _xklLastErrorCode = Success;
 
-  retCode = XGrabKey( _xklDpy, keyCode, modifiers, _xklRootWindow,
+  retCode = XGrabKey( _xklDpy, keycode, modifiers, _xklRootWindow,
                       True, GrabModeAsync, GrabModeAsync );
   XSync( _xklDpy, False );
 
-  XklDebug( 100, "trying to listen: %d/%d\n", retCode, _xklLastErrorCode );
+  XklDebug( 100, "XGrabKey recode %d/error %d\n", retCode, _xklLastErrorCode );
 
   retCode = ( _xklLastErrorCode == Success );
 
@@ -187,16 +308,12 @@ Bool XklGrabKey( int key, unsigned modifiers )
   return retCode;
 }
 
-Bool XklUngrabKey( int key, unsigned modifiers )
+Bool XklUngrabKey( int keycode, unsigned modifiers )
 {
-  int keyCode;
-
-  keyCode = XKeysymToKeycode( _xklDpy, key );
-
-  if( ( KeyCode ) NULL == keyCode )
+  if( ( KeyCode ) NULL == keycode )
     return False;
 
-  return Success == XUngrabKey( _xklDpy, keyCode, 0, _xklRootWindow );
+  return Success == XUngrabKey( _xklDpy, keycode, 0, _xklRootWindow );
 }
 
 int XklGetNextGroup(  )
@@ -238,7 +355,7 @@ void XklSetTransparent( Window win, Bool transparent )
   {
     XklDebug( 150, "No app window!\n" );
     appWin = win;
-//    return;
+/*    return; */
   }
 
   wasTransparent = XklIsTransparent( appWin );
@@ -315,7 +432,7 @@ void _XklAddAppWindow( Window appWin, Window parent, Bool ignoreExistingState,
     if( _xklCurClient == appWin )
     {
       if( ( _xklSecondaryGroupsMask & ( 1 << defGroupToUse ) ) != 0 )
-        XklAllowOneSwitchToSecondaryGroup(  );
+        XklAllowOneSwitchToSecondaryGroup();
       XklLockGroup( defGroupToUse );
     }
   }
@@ -442,9 +559,10 @@ Bool _XklLoadWindowTree(  )
 {
   Window focused;
   int revert;
-  Bool retval, haveAppWindow;
+  Bool retval = True, haveAppWindow;
 
-  retval = _XklLoadSubtree( _xklRootWindow, 0, &_xklCurState );
+  if( _xklListenerType & XKLL_MANAGE_WINDOW_STATES )
+    retval = _XklLoadSubtree( _xklRootWindow, 0, &_xklCurState );
 
   XGetInputFocus( _xklDpy, &focused, &revert );
 
@@ -512,7 +630,7 @@ Bool _XklGetAppState( Window appWin, XklState * state_return )
         ( _xklDpy, appWin, _xklAtoms[XKLAVIER_STATE], 0L,
           XKLAVIER_STATE_PROP_LENGTH, False,
           XA_INTEGER, &type_ret, &format_ret, &nitems, &rest,
-          ( unsigned char ** ) &prop ) == Success )
+          ( unsigned char ** ) ( void * ) &prop ) == Success )
       && ( type_ret == XA_INTEGER ) && ( format_ret == 32 ) )
   {
     grp = prop[0];
@@ -578,7 +696,7 @@ void _XklSelectInput( Window win, long mask )
 {
   if( _xklRootWindow == win )
     XklDebug( 160,
-              "Someone is looking for " WINID_FORMAT " on root window ***\n",
+              "Someone is looking for %lx on root window ***\n",
               mask );
 
   XSelectInput( _xklDpy, win, mask );
@@ -606,22 +724,22 @@ void _XklTryCallStateCallback( XklStateChange changeType,
   XklDebug( 150,
             "changeType: %d, group: %d, secondaryGroupMask: %X, allowsecondary: %d\n",
             changeType, group, _xklSecondaryGroupsMask,
-            _xklAllowSecondaryGroupOnce );
+            _XklIsOneSwitchToSecondaryGroupAllowed() );
 
   if( changeType == GROUP_CHANGED )
   {
     if( !restore )
     {
       if( ( _xklSecondaryGroupsMask & ( 1 << group ) ) != 0 &&
-          !_xklAllowSecondaryGroupOnce )
+          !_XklIsOneSwitchToSecondaryGroupAllowed() )
       {
         XklDebug( 150, "secondary -> go next\n" );
         group = XklGetNextGroup(  );
         XklLockGroup( group );
-        return;                 // we do not need to revalidate
+        return;                 /* we do not need to revalidate */
       }
     }
-    _xklAllowSecondaryGroupOnce = False;
+    _XklOneSwitchToSecondaryGroupPerformed();
   }
   if( stateCallback != NULL )
   {
@@ -637,11 +755,10 @@ Bool _XklIsTransparentAppWindow( Window appWin )
   int format_ret;
   unsigned long nitems, rest;
   CARD32 *prop = NULL;
-  Bool retVal;
   if( ( XGetWindowProperty
         ( _xklDpy, appWin, _xklAtoms[XKLAVIER_TRANSPARENT], 0L, 1, False,
           XA_INTEGER, &type_ret, &format_ret, &nitems, &rest,
-          ( unsigned char ** ) &prop ) == Success )
+          ( unsigned char ** ) ( void * ) &prop ) == Success )
       && ( type_ret == XA_INTEGER ) && ( format_ret == 32 ) )
   {
     if( prop != NULL )
@@ -650,3 +767,95 @@ Bool _XklIsTransparentAppWindow( Window appWin )
   }
   return False;
 }
+
+void _XklEnsureVTableInited( void )
+{
+  if ( xklVTable == NULL )
+  {
+    XklDebug( 0, "ERROR: XKL VTable is NOT initialized.\n" );
+    /* force the crash! */
+    char *p = NULL; *p = '\0';
+  }
+}
+
+const char *XklGetBackendName( void )
+{
+  return xklVTable->id;
+}
+
+int XklGetBackendFeatures( void )
+{
+  return xklVTable->features;
+}
+
+void _XklResetAllInfo( const char reason[] )
+{
+  XklDebug( 150, "Resetting all the cached info, reason: [%s]\n", reason );
+  _XklEnsureVTableInited();
+  if( !(*xklVTable->xklIfCachedInfoEqualsActualHandler)() )
+  {
+    (*xklVTable->xklFreeAllInfoHandler)();
+    (*xklVTable->xklLoadAllInfoHandler)();
+  } else
+    XklDebug( 100, "NOT Resetting the cache: same configuration\n" );
+}
+
+/**
+ * Calling through vtable
+ */
+const char **XklGetGroupNames( void )
+{
+  _XklEnsureVTableInited();
+  return (*xklVTable->xklGetGroupNamesHandler)();
+}
+
+unsigned XklGetNumGroups( void )
+{
+  _XklEnsureVTableInited();
+  return (*xklVTable->xklGetNumGroupsHandler)();
+}
+
+void XklLockGroup( int group )
+{
+  _XklEnsureVTableInited();
+  (*xklVTable->xklLockGroupHandler)( group );
+}
+
+int XklPauseListen( void )
+{
+  _XklEnsureVTableInited();
+  return (*xklVTable->xklPauseListenHandler)();
+}
+
+int XklResumeListen( void )
+{
+  _XklEnsureVTableInited();
+  XklDebug( 150, "listenerType: %x\n", _xklListenerType );
+  if( (*xklVTable->xklResumeListenHandler)() )
+    return 1;
+  
+  _XklSelectInputMerging( _xklRootWindow,
+                          SubstructureNotifyMask | PropertyChangeMask );
+  _XklEnsureVTableInited();
+  (*xklVTable->xklGetRealStateHandler)( &_xklCurState );
+  return 0;
+}
+
+Bool _XklLoadAllInfo( void )
+{
+  _XklEnsureVTableInited();
+  return (*xklVTable->xklLoadAllInfoHandler)();
+}
+
+void _XklFreeAllInfo( void )
+{
+  _XklEnsureVTableInited();
+  (*xklVTable->xklFreeAllInfoHandler)();
+}
+
+unsigned XklGetMaxNumGroups( void )
+{
+  _XklEnsureVTableInited();
+  return (*xklVTable->xklGetMaxNumGroupsHandler)();
+}
+
